#include <irrlicht.h>
#define EPS 0.00001

using namespace irr;
using namespace core;
using namespace scene;
using namespace video;

// Calcula o mdulo de uma coordenada

irr::f32 absf32(irr::f32 in) {
    irr::f32 out;
    out = in;
    if(in < 0.0)
        out = in * -1.0;
    return out;
}

// Decide se duas linhas se intersectam

int LineLineIntersect(
   core::vector3df p1,core::vector3df p2,core::vector3df p3,
   core::vector3df p4,core::vector3df *pa,core::vector3df *pb,
   float *mua, float *mub)
{
   core::vector3df p13,p43,p21;
   float d1343,d4321,d1321,d4343,d2121;
   float numer,denom;

   p13.X = p1.X - p3.X;
   p13.Y = p1.Y - p3.Y;
   p13.Z = p1.Z - p3.Z;
   p43.X = p4.X - p3.X;
   p43.Y = p4.Y - p3.Y;
   p43.Z = p4.Z - p3.Z;
   if (absf32(p43.X)  < EPS && absf32(p43.Y)  < EPS && absf32(p43.Z)  < EPS)
      return(0);
   p21.X = p2.X - p1.X;
   p21.Y = p2.Y - p1.Y;
   p21.Z = p2.Z - p1.Z;
   if (absf32(p21.X)  < EPS && absf32(p21.Y)  < EPS && absf32(p21.Z)  < EPS)
      return(0);

   d1343 = p13.X * p43.X + p13.Y * p43.Y + p13.Z * p43.Z;
   d4321 = p43.X * p21.X + p43.Y * p21.Y + p43.Z * p21.Z;
   d1321 = p13.X * p21.X + p13.Y * p21.Y + p13.Z * p21.Z;
   d4343 = p43.X * p43.X + p43.Y * p43.Y + p43.Z * p43.Z;
   d2121 = p21.X * p21.X + p21.Y * p21.Y + p21.Z * p21.Z;

   denom = d2121 * d4343 - d4321 * d4321;
   if (absf32(denom) < EPS)
      return(0);
   numer = d1343 * d4321 - d1321 * d4343;

   *mua = numer / denom;
   *mub = (d1343 + d4321 * (*mua)) / d4343;

   pa->X = p1.X + *mua * p21.X;
   pa->Y = p1.Y + *mua * p21.Y;
   pa->Z = p1.Z + *mua * p21.Z;
   pb->X = p3.X + *mub * p43.X;
   pb->Y = p3.Y + *mub * p43.Y;
   pb->Z = p3.Z + *mub * p43.Z;

   return(1);
}

// Decide se o portal cabe em um tringulo na direo de uma linha

bool portalFitsTriangleInDirection(core::line3d<f32> lineDirection, core::vector3df intersection, core::triangle3df tri, 
        core::vector3df diretor_vertical, core::vector3df diretor_horizontal, core::vector3df *limit) {
    
    core::vector3df pointA;
    core::vector3df pointB;
    core::vector3df pointC;
    
    core::vector3df trinormal;
    core::vector3df intersecA;
    core::vector3df intersecB;
    irr::f32 mua, mub;
    
    pointA=tri.pointA;
    pointB=tri.pointB;
    pointC=tri.pointC;
    
    core::line3d<f32> line1(pointA, pointB); 
    core::line3d<f32> line2(pointB, pointC);
    core::line3d<f32> line3(pointA, pointC);
      
    trinormal=tri.getNormal();
    trinormal=trinormal.normalize();
    
    core::line3d<f32> lines[3];
    
    lines[0] = line1;
    lines[1] = line2;
    lines[2] = line3;
    
    for(int i = 0; i < 3; i++)
    {
        if(LineLineIntersect(lineDirection.start, lineDirection.end,
           lines[i].start,lines[i].end,
           &intersecA,&intersecB, 
           &mua,&mub)) {
            if(lineDirection.isPointBetweenStartAndEnd(intersecB)) {
                *limit = intersecB;
                return false;
            }
        }
    }
    return true;
}

/* Decide se um portal cabe em uma parede desejada. O processo executado por
este algoritmo :
    * Chamamos de T o tringulo onde houve a coliso.
    * A partir centro do portal, criamos quatro linhas, cada uma ligando o centro aos pontos onde ficaro os vrtices do 
portal.
    * Para cada linha, verificamos se a mesma est contida em T.
    * Caso uma linha esteja contida em T, isto significa que o portal cabe na parede na direo daquela linha. 
Caso contrrio, achamos o prximo tringulo T' da parede na direo da linha.
    * Caso no haja nenhum, ou T' no esteja no mesmo plano de T, o algoritmo termina e descobrimos que o portal no cabe 
naquela parede.
    * Caso contrrio, continuamos o algoritmo com T sendo o novo tringulo encontrado.
    * Decidimos que o portal cabe na parede quando as 4 linhas tiverem sido exploradas at o final. */


bool portalFitsWall(core::vector3df intersection, core::triangle3df tri, core::vector3df diretor_vertical,
        core::vector3df diretor_horizontal, int width, int height,video::IVideoDriver* driver, 
        scene::ISceneManager* smgr, scene::ITriangleSelector* selector) {
    
    core::line3d<f32> lineDownLeft,lineDownRight,lineUpLeft,lineUpRight, lineBusca;
    
    core::vector3df limit, vectorDownLeft, vectorDownRight, vectorUpLeft, vectorUpRight,
    intersectionDownLeft, intersectionDownRight, intersectionUpLeft, intersectionUpRight,
    trinormal, trinormalNova;
    
    core::triangle3df triDownLeft, triDownRight, triUpLeft, triUpRight, oldTri;
    float faltaDownLeft, faltaDownRight, faltaUpLeft, faltaUpRight;
    
    triDownLeft = triDownRight = triUpLeft = triUpRight = tri;
    intersectionDownLeft = intersectionDownRight = intersectionUpLeft = intersectionUpRight = intersection;
    
    lineDownLeft.start = intersection;
    lineDownLeft.end = intersection + diretor_horizontal * height/2 + diretor_vertical * width/2;
    faltaDownLeft = lineDownLeft.getLength();
    vectorDownLeft = lineDownLeft.getVector().normalize();

    lineDownRight.start = intersection;
    lineDownRight.end = intersection + diretor_horizontal * height/2 - diretor_vertical * width/2;
    faltaDownRight = lineDownRight.getLength();
    vectorDownRight = lineDownRight.getVector().normalize();

    lineUpLeft.start = intersection;
    lineUpLeft.end = intersection - diretor_horizontal * height/2 + diretor_vertical * width/2;
    faltaUpLeft = lineUpLeft.getLength();
    vectorUpLeft = lineUpLeft.getVector().normalize();

    lineUpRight.start = intersection;
    lineUpRight.end = intersection - diretor_horizontal * height/2 - diretor_vertical * width/2;
    faltaUpRight = lineUpRight.getLength();
    vectorUpRight = lineUpRight.getVector().normalize();
    
    while(faltaDownLeft > 0 || faltaDownRight > 0 || faltaUpLeft > 0 || faltaUpRight > 0) {
        if(faltaDownLeft > 0) {        
            if(!portalFitsTriangleInDirection(lineDownLeft, intersectionDownLeft,triDownLeft,diretor_vertical, diretor_horizontal, &limit)){
                lineDownLeft.end = limit;
                faltaDownLeft -=  lineDownLeft.getLength();
                
                trinormal=triDownLeft.getNormal();
                trinormal=trinormal.normalize();
                
                lineBusca.start = limit + trinormal * 5 + vectorDownLeft * 1;
                lineBusca.end = limit - trinormal * 5 + vectorDownLeft * 1;
                
                oldTri = triDownLeft;
                
                if(!smgr->getSceneCollisionManager()->getCollisionPoint(lineBusca, selector, intersectionDownLeft, triDownLeft)) {
                    return false;
                }

                
                trinormalNova=triDownLeft.getNormal();
                trinormalNova=trinormalNova.normalize();
                
                if(trinormalNova != trinormal) {
                    return false;
                }
                    
                faltaDownLeft -= 1; //-1 pois j andei 1 na direo do novo tringulo
                
                lineDownLeft.start = intersectionDownLeft;
                lineDownLeft.end = intersectionDownLeft + vectorDownLeft * faltaDownLeft; 
                
            }
            else
                faltaDownLeft = 0;
        }
        if(faltaDownRight > 0) {        
            if(!portalFitsTriangleInDirection(lineDownRight, intersectionDownRight,triDownRight,diretor_vertical, diretor_horizontal, &limit)){
                lineDownRight.end = limit;
                faltaDownRight -=  lineDownRight.getLength();
                
                trinormal=triDownRight.getNormal();
                trinormal=trinormal.normalize();
                
                lineBusca.start = limit + trinormal * 5 + vectorDownRight * 1;
                lineBusca.end = limit - trinormal * 5 + vectorDownRight * 1;
                
                oldTri = triDownRight;
                
                if(!smgr->getSceneCollisionManager()->getCollisionPoint(lineBusca, selector, intersectionDownRight, triDownRight)) {
                    return false;
                }

                
                trinormalNova=triDownRight.getNormal();
                trinormalNova=trinormalNova.normalize();
                
                if(trinormalNova != trinormal) {
                    return false;
                }
                    
                faltaDownRight -= 1; //-1 pois j andei 1 na direo do novo tringulo
                
                lineDownRight.start = intersectionDownRight;
                lineDownRight.end = intersectionDownRight + vectorDownRight * faltaDownRight; 
                
            }
            else
                faltaDownRight = 0;
        }
        if(faltaUpLeft > 0) {        
            if(!portalFitsTriangleInDirection(lineUpLeft, intersectionUpLeft,triUpLeft,diretor_vertical, diretor_horizontal, &limit)){
                lineUpLeft.end = limit;
                faltaUpLeft -=  lineUpLeft.getLength();
                
                trinormal=triUpLeft.getNormal();
                trinormal=trinormal.normalize();
                
                lineBusca.start = limit + trinormal * 5 + vectorUpLeft * 1;
                lineBusca.end = limit - trinormal * 5 + vectorUpLeft * 1;
                
                oldTri = triUpLeft;
                
                if(!smgr->getSceneCollisionManager()->getCollisionPoint(lineBusca, selector, intersectionUpLeft, triUpLeft)) {
                    return false;
                }
                
                trinormalNova=triUpLeft.getNormal();
                trinormalNova=trinormalNova.normalize();
                
                if(trinormalNova != trinormal) {
                    return false;
                }
                    
                
                faltaUpLeft -= 1; //-1 pois j andei 1 na direo do novo tringulo
                
                lineUpLeft.start = intersectionUpLeft;
                lineUpLeft.end = intersectionUpLeft + vectorUpLeft * faltaUpLeft; 
                
            }
            else
                faltaUpLeft = 0;
        }
        if(faltaUpRight > 0) {        
            if(!portalFitsTriangleInDirection(lineUpRight, intersectionUpRight,triUpRight,diretor_vertical, diretor_horizontal, &limit)){
                lineUpRight.end = limit;
                faltaUpRight -=  lineUpRight.getLength();
                
                trinormal=triUpRight.getNormal();
                trinormal=trinormal.normalize();
                
                lineBusca.start = limit + trinormal * 5 + vectorUpRight * 1;
                lineBusca.end = limit - trinormal * 5 + vectorUpRight * 1;
                
                oldTri = triUpRight;
                
                if(!smgr->getSceneCollisionManager()->getCollisionPoint(lineBusca, selector, intersectionUpRight, triUpRight)) {
                    return false;
                }
                
                trinormalNova=triUpRight.getNormal();
                trinormalNova=trinormalNova.normalize();
                
                if(trinormalNova != trinormal) {
                    return false;
                }
                    
                faltaUpRight -= 1; //-1 pois j andei 1 na direo do novo tringulo
                
                lineUpRight.start = intersectionUpRight;
                lineUpRight.end = intersectionUpRight + vectorUpRight * faltaUpRight; 
                
            }
            else
                faltaUpRight = 0;
        }
    }
    return true;
}
